@@ -1,6 +1,6 @@
 package Audio::StreamGenerator;
 
-our $VERSION = 0.05;
+our $VERSION = 0.06;
 
 use strict;
 use warnings;
@@ -47,9 +47,10 @@ sub new {
 
 sub stream {
     my $self = shift;
+    my @buffer;
 
     $self->{source} = $self->_do_get_new_source();
-    $self->{buffer} = [];
+    $self->{buffer} = \@buffer;
     $self->{skip}   = 0;
 
     my $short_clips_seen = 0;
@@ -58,13 +59,17 @@ sub stream {
     my @channels;
     push @channels, $_ for 0 ... ( $self->{channels_amount} - 1 );
 
+    my $eof = undef;
+
     while (1) {
 
-        if ( eof( $self->{source} ) || $self->{skip} ) {
+        $eof = eof( $self->{source} ) unless defined $eof;
+        if ( $eof || $self->{skip} ) {
+            $eof = undef;
 
             if ( $self->{skip} ) {
                 $logger->info('shortening buffer for skip...');
-                pop @{ $self->{buffer} }
+                pop @buffer
                     for 0 ... ( $self->{sample_rate} * ( $self->{normal_fade_seconds} - $self->{skip_fade_seconds} ) );
             }
 
@@ -76,6 +81,7 @@ sub stream {
             if ( $old_elapsed_seconds < ( $self->{normal_fade_seconds} * 2 ) ) {
                 $short_clips_seen++;
                 if ( $short_clips_seen >= 2 ) {
+                    $self->{skip} = 0;
                     $logger->info('not mixing');
                     next;
                 } else {
@@ -88,11 +94,14 @@ sub stream {
                 $logger->info('mixing');
             }
 
+            # make the buffer mixable
+            @buffer = map { $self->_unpack_sample($_) } @buffer;
+
             my $index                  = 0;
             my $last_loud_sample_index = -1;
             my $threshold              = $maxint * $self->{max_vol_before_mix_fraction};
             my $max_old                = 0;
-            foreach my $sample ( @{ $self->{buffer} } ) {
+            foreach my $sample ( @buffer ) {
                 foreach (@channels) {
                     my $single_sample = $sample->[$_];
                     $single_sample *= -1 if $single_sample < 0;
@@ -110,16 +119,16 @@ sub stream {
             $logger->info("loudest sample value: $max_old");
 
             my @new_buffer;
-            while ( @new_buffer < @{ $self->{buffer} } ) {
-                my $sample = $self->_get_sample();
+            while ( @new_buffer < @buffer ) {
+                my $sample = $self->_unpack_sample($self->_get_sample);
                 last if !defined($sample);
                 push( @new_buffer, $sample );
             }
 
             my @max   = (0) x $self->{channels_amount};
-            my $total = scalar( @{ $self->{buffer} } );
+            my $total = @buffer;
             $index = -1;
-            foreach my $sample ( @{ $self->{buffer} } ) {
+            foreach my $sample (@buffer) {
                 $index++;
                 my $togo = $total - $index;
 
@@ -165,42 +174,36 @@ sub stream {
                 }
             }
 
-            push( @{ $self->{buffer} }, @new_buffer );
+            push( @buffer, @new_buffer );
 
-            my $channel = 0;
+            # Volume adjustment
+            foreach my $channel ( grep { $max[$_] > $maxint } @channels ) {
+                $logger->info("channel $channel needs volume adjustment");
 
-            foreach my $channel (@channels) {
-                $logger->info("channel $channel needs volume adjustment")
-                    if ( $max[$channel] > $maxint );
-            }
-
-            foreach my $sample ( @{ $self->{buffer} } ) {
-                for my $channel (@channels) {
-                    if ( $max[$channel] > $maxint ) {
-                        $sample->[$channel] =
-                            ( $sample->[$channel] / $max[$channel] ) * $maxint;
-                    }
+                foreach my $sample ( @buffer ) {
+                    $sample->[$channel] =
+                        ( $sample->[$channel] / $max[$channel] ) * $maxint;
                 }
             }
 
             $self->{skip} = 0;
-
         }
 
-        while ( @{ $self->{buffer} } < ( $self->{normal_fade_seconds} * $self->{sample_rate} ) ) {
+        my $needed = $self->{normal_fade_seconds} * $self->{sample_rate};
+        while ( @buffer < $needed ) {
             my $sample = $self->_get_sample();
-            last if !defined($sample);
-            push( @{ $self->{buffer} }, $sample );
+            if (!defined($sample)) {
+                $eof = 1;
+                last;
+            }
+            push( @buffer, $sample );
         }
 
-        $self->_send_one_sample();
+        $self->_send_one_sample() while @buffer >= $needed;
 
-        if ( !( $self->{elapsed} % $self->{sample_rate} )
-            && defined( $self->{run_every_second} ) )
-        {
+        if ( defined( $self->{run_every_second} ) && !( $self->{elapsed} % $self->{sample_rate} )) {
             $self->{run_every_second}($self);
         }
-
     }
 
 }
@@ -219,27 +222,38 @@ sub _send_one_sample {
     my $self   = shift;
     my $sample = shift @{ $self->{buffer} };
     my $fh     = $self->{out_fh};
-    print $fh map { pack 's*', $_ } @$sample;
+
+    if ( ref $sample eq 'ARRAY' ) {
+        print $fh map { pack 's*', $_ } @$sample;
+    } else {
+        print $fh $sample;
+    }
+}
+
+sub _unpack_sample {
+    my ($self, $sample) = @_;
+
+    return $sample unless defined $sample;
+    return $sample if ref $sample eq 'ARRAY';
+    return [unpack 's*', $sample];
 }
 
 sub _get_sample {
     my $self = shift;
-    return undef if eof( $self->{source} );
+
     my $data;
-    read( $self->{source}, $data, $self->{channels_amount} * 2 );
+    my $bytes = $self->{channels_amount} * 2;
+    my $len   = read( $self->{source}, $data, $bytes );
+
+    return undef if $len == 0;
     $self->{elapsed}++;
 
-    if ( length($data) == ( $self->{channels_amount} * 2 ) ) {
-        my @sample;
-        while ( length($data) ) {
-            my $bytes_this_sample = substr( $data, 0, $self->{channels_amount} * 2, '' );
-            push( @sample, unpack 's*', $bytes_this_sample );
-        }
-        return \@sample;
-    } else {
-        my @sample = (0) x ( $self->{channels_amount} * 2 );
-        return \@sample;
+    if ( $len != $bytes ) {
+        # pack's "s" is 16 bit unsigned, so we need two bytes
+        $data = "\x00\x00" x $bytes ;
     }
+
+    return $data;
 }
 
 sub _do_get_new_source {
@@ -278,7 +292,7 @@ Audio::StreamGenerator - create a 'radio' stream by mixing ('cross fading') mult
 
     my $out_fh;
     open ($out_fh, '|-', $out_command);
-    
+
     sub get_new_source {
         my $fullpath = '/path/to/some/audiofile.flac';
         my @ffmpeg_cmd = (
@@ -295,7 +309,7 @@ Audio::StreamGenerator - create a 'radio' stream by mixing ('cross fading') mult
         open(my $source, '-|', @ffmpeg_cmd);
         return $source;
     }
-    
+
     sub run_every_second {
         my $streamert = shift;
         my $position = $streamert->get_elapsed_seconds();
@@ -304,30 +318,30 @@ Audio::StreamGenerator - create a 'radio' stream by mixing ('cross fading') mult
             $streamert->skip()
         }
     }
-    
+
     my $streamer = Audio::StreamGenerator->new(
         out_fh => $out_fh,
         get_new_source => \&get_new_source,
         run_every_second => \&run_every_second,
     );
-    
+
     $streamer->stream();
 
 =head1 DESCRIPTION
 
-This module creates a 'live' audio stream that can be broadcast using streaming technologies like Icecast or HTTP Live Streaming. 
+This module creates a 'live' audio stream that can be broadcast using streaming technologies like Icecast or HTTP Live Streaming.
 
-It creates one ongoing audio stream by mixing or 'crossfading' multiple sources (normally audio files). 
+It creates one ongoing audio stream by mixing or 'crossfading' multiple sources (normally audio files).
 
-Although there is nothing stopping you from using this to generate a file that can be played back later, its intended use is to create a 'radio' stream that can be streamed or 'broadcast' live on the internet. 
+Although there is nothing stopping you from using this to generate a file that can be played back later, its intended use is to create a 'radio' stream that can be streamed or 'broadcast' live on the internet.
 
-The module takes raw PCM audio from a file handle as input, and outputs raw PCM audio to another file handle. This means that an external program is necessary to decode (mp3/flac/etc) source files, and to encode & stream the actual output. For both purposes, ffmpeg is recommended - but anything that can produce and/or receive raw PCM audio should do. 
+The module takes raw PCM audio from a file handle as input, and outputs raw PCM audio to another file handle. This means that an external program is necessary to decode (mp3/flac/etc) source files, and to encode & stream the actual output. For both purposes, ffmpeg is recommended - but anything that can produce and/or receive raw PCM audio should do.
 
 =head1 CONSTRUCTOR METHOD
 
     my $streamer = Audio::StreamGenerator->new( %options );
 
-Creates a new StreamGenerator object and returns it. 
+Creates a new StreamGenerator object and returns it.
 
 =head1 OPTIONS
 
@@ -346,16 +360,16 @@ The following options can be specified:
 
 =head2 out_fh
 
-The outgoing file handle - this is where the generated signed 16-bit little-endian PCM audio stream is sent to. 
+The outgoing file handle - this is where the generated signed 16-bit little-endian PCM audio stream is sent to.
 
-Note that StreamGenerator has no notion of time - if you don't slow it down, it will process data as fast as it can - which is faster than your listeners are able to play the stream. 
-On Icecast, this will cause listeners to be disconnected because they are "too far behind". 
+Note that StreamGenerator has no notion of time - if you don't slow it down, it will process data as fast as it can - which is faster than your listeners are able to play the stream.
+On Icecast, this will cause listeners to be disconnected because they are "too far behind".
 
-This can be addressed by making sure that the out_fh process consumes the audio no faster than realtime. 
+This can be addressed by making sure that the out_fh process consumes the audio no faster than realtime.
 
-If you are using ffmpeg, you can achieve this with its '-re' option. 
+If you are using ffmpeg, you can achieve this with its '-re' option.
 
-Another possibility is to first pipe the data to a command like 'pv' to rate limit the data. An additional advantage of 'pv' is that it can also add a buffer between the StreamGenerator and the encoder, which can absorb any short delays that may occur when StreamGenerator is switching to a new track. 
+Another possibility is to first pipe the data to a command like 'pv' to rate limit the data. An additional advantage of 'pv' is that it can also add a buffer between the StreamGenerator and the encoder, which can absorb any short delays that may occur when StreamGenerator is switching to a new track.
 
 Example:
 
@@ -363,36 +377,36 @@ Example:
 
 This will tell pv to be quiet (no output to STDERR), to allow a maximum throughput of 44100 samples per second * 2 bytes per sample * 2 channels = 176400 bytes per second, and keep a buffer of 176400 Bps * 20 seconds = 3528000 bytes
 
-The out_fh command is also the place where you could insert a sound processing solution like the command line version of L<Stereo tool|https://www.stereotool.com/> - just pipe the audio first to that tool, and from there to your encoder. 
+The out_fh command is also the place where you could insert a sound processing solution like the command line version of L<Stereo tool|https://www.stereotool.com/> - just pipe the audio first to that tool, and from there to your encoder.
 
 =head2 get_new_source
 
-Reference to a sub that will be called every time that a new source (audio file) is needed. Needs to return a readable filehandle that will output signed 16-bit little-endian PCM audio. 
-    
+Reference to a sub that will be called every time that a new source (audio file) is needed. Needs to return a readable filehandle that will output signed 16-bit little-endian PCM audio.
+
 =head2 run_every_second
 
-This sub will be run after each second of playback, with the StreamGenerator object as an argument. This can be used to do things like updating a UI with the current playing position - or to call the skip() method if we need to skip to the next source. 
-    
+This sub will be run after each second of playback, with the StreamGenerator object as an argument. This can be used to do things like updating a UI with the current playing position - or to call the skip() method if we need to skip to the next source.
+
 =head2 normal_fade_seconds
 
 Amount of seconds that we want tracks to overlap. This is only the initial/max value - the mixing algorithm may decide to mix less seconds if the old track ends with loud samples.
-    
+
 =head2 skip_fade_seconds
 
-When 'skipping' to the next song using the skip() method (for example, after a user clicked a "next song" button on some web interface), we mix less seconds than normally, simply because mixing 5+ seconds in the middle of the old track sounds pretty bad. This value has to be lower than normal_fade_seconds. 
-    
+When 'skipping' to the next song using the skip() method (for example, after a user clicked a "next song" button on some web interface), we mix less seconds than normally, simply because mixing 5+ seconds in the middle of the old track sounds pretty bad. This value has to be lower than normal_fade_seconds.
+
 =head2 sample_rate
 
-The amount of samples per second (both incoming & outgoing), normally this is 44100 for standard CD-quality audio. 
-    
+The amount of samples per second (both incoming & outgoing), normally this is 44100 for standard CD-quality audio.
+
 =head2 channels_amount
 
-Amount of audio channels, this is normally 2 (stereo). 
+Amount of audio channels, this is normally 2 (stereo).
 
 =head2 max_vol_before_mix_fraction
 
-This tells StreamGenerator what the minimum volume of a 'loud' sample is. It is expressed as a fraction of the maximum volume. 
-When mixing 2 tracks, StreamGenerator needs to find out what the last loud sample of the old track is so that it can start the next song immediately after that. 
+This tells StreamGenerator what the minimum volume of a 'loud' sample is. It is expressed as a fraction of the maximum volume.
+When mixing 2 tracks, StreamGenerator needs to find out what the last loud sample of the old track is so that it can start the next song immediately after that.
 
 =head1 METHODS
 
@@ -406,18 +420,18 @@ Start the actual audio stream.
 
     $streamer->skip();
 
-Skip to the next track without finishing the current one. This can be called from the "run_every_second" sub, for example after checking whether a 'skip' flag was set in a database, or whether a file exists. 
+Skip to the next track without finishing the current one. This can be called from the "run_every_second" sub, for example after checking whether a 'skip' flag was set in a database, or whether a file exists.
 
 =head2 get_elapsed_samples
 
     my $elapsed_samples = $streamer->get_elapsed_samples();
     print "$elapsed_samples played so far\r";
 
-Get the amount of played samples in the current track - this can be called from the "run_every_second" sub. 
+Get the amount of played samples in the current track - this can be called from the "run_every_second" sub.
 
 =head2 get_elapsed_seconds
 
     my $elapsed_seconds = $streamer->get_elapsed_seconds();
     print "now at position $elapsed_seconds of the current track\r";
 
-Get the amount of elapsed seconds in the current track - in other words the current position in the track. This equals to get_elapsed_samples/sample_rate . 
+Get the amount of elapsed seconds in the current track - in other words the current position in the track. This equals to get_elapsed_samples/sample_rate .
